import pt.tumba.parser.HTMLParser;

import java.io.*;
import java.util.*;

public class DmozInfoExtractor {

    private static final String CONFIG_FILE = "config.properties";

    private String languageProfile;
    private String categoryFile;

    public DmozInfoExtractor(String languageProfile, String categoryFile) {
        this.languageProfile = languageProfile;
        this.categoryFile = categoryFile;
    }

    public String getLanguageProfile() {
        return languageProfile;
    }

    public void setLanguageProfile(String languageProfile) {
        this.languageProfile = languageProfile;
    }

    public String getCategoryFile() {
        return categoryFile;
    }

    public void setCategoryFile(String categoryFile) {
        this.categoryFile = categoryFile;
    }

    public void start() {
        try {
            Properties systemSettings = System.getProperties();
            systemSettings.load(
                    DmozInfoExtractor.class.getResourceAsStream(
                            CONFIG_FILE));
            System.setProperties(systemSettings);
            File fin = new File(categoryFile);
            File fout = new File(categoryFile + ".keys");
            Map<Keyword, Keyword> keywordMap = new HashMap<Keyword, Keyword>(1000);
            if (fin.exists()) {
                HTMLParser htmlParser = new HTMLParser(languageProfile);
                FileInputStream fstream = new FileInputStream(fin);
                DataInputStream in = new DataInputStream(fstream);
                BufferedReader br = new BufferedReader(new InputStreamReader(in));
                String line;
                int lineCount = 0;
                while ((line = br.readLine()) != null) {
                    try {
                        lineCount++;
                        String[] cols = line.split("\\|");
                        String url = cols[2];
                        ConvertUtils.initParser(url, htmlParser);
                        htmlParser.processData();
                        if (!htmlParser.getMetaData().getKeywords().equals("")) {
                            String keywords = htmlParser.getMetaData().getKeywords();
                            String[] keywordList = keywords.split(",");
                            for (String word : keywordList) {
                                Keyword keyword = new Keyword(word.trim());
                                if (!keywordMap.containsKey(keyword)) {
                                    keyword.setCount(1);
                                    keywordMap.put(keyword, keyword);
                                } else {
                                    keyword = keywordMap.get(keyword);
                                    keyword.setCount(keyword.getCount() + 1);
                                }
                            }
                        }
                    } catch (Exception e) {
                        System.out.println("Error processing line : " + line);
                        System.out.println(e.getMessage());
                        e.printStackTrace();
                    }

                    if (lineCount % 10 == 0) {
                        // Every 100 line then write to file
                        FileOutputStream foutStream = new FileOutputStream(fout);
                        DataOutputStream out = new DataOutputStream(foutStream);
                        BufferedWriter bw = new BufferedWriter(new OutputStreamWriter(out));
                        Collection<Keyword> col = keywordMap.values();
                        ArrayList<Keyword> aList = new ArrayList<Keyword>(col);
                        Collections.sort(aList);
                        for (Keyword k : aList) {
                            bw.write(k.getValue() + ":" + k.getCount() + "\n");
                        }
                        bw.write("Line count:----" + lineCount);
                        bw.close();
                    }
                }
                in.close();
            } else {
                System.out.println("Category file does not exist.");
                return;
            }
        } catch (Exception e) {
            System.err.println(e.getMessage());
            e.printStackTrace();
        }

    }

    /**
     * The main method of this class, parsing parameters from the command line, invoquing
     * the parser to process the given input, and dumping the results in RDF.
     *
     * @param args The command line arguments
     */
    public static void main(String args[]) throws Exception {
        if (args.length != 2) {
            System.out.println("Usage: java -jar extractor.jar <language profile folder> <categories file>");
            return;
        }
        DmozInfoExtractor extractor = new DmozInfoExtractor(args[0], args[1]);
        extractor.start();
    }

}
